@@ -1,7 +1,7 @@
 /**
  * @file item_list_view.c  presenting items in a GtkTreeView
  *  
- * Copyright (C) 2004-2011 Lars Lindner <lars.lindner@gmail.com>
+ * Copyright (C) 2004-2012 Lars Windolf <lars.lindner@gmail.com>
  * Copyright (C) 2004-2006 Nathan J. Conrad <t98502@users.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -21,6 +21,10 @@
 
 #include "ui/item_list_view.h"
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <string.h>
 #include <glib.h>
 #include <gdk/gdkkeysyms.h>
@@ -64,7 +68,7 @@ enum is_columns {
 	IS_PARENT,		/**< Parent node pointer */
 	IS_FAVICON,		/**< Pixbuf reference to the item's feed's icon */
 	IS_ENCICON,		/**< Pixbuf reference to the item's enclosure icon */
-	IS_ENCLOSURE,		/**< Flag wether enclosure is attached or not */
+	IS_ENCLOSURE,		/**< Flag whether enclosure is attached or not */
 	IS_SOURCE,		/**< Source node pointer */
 	IS_STATE,		/**< Original item state (unread, flagged...) for sorting */
 	ITEMSTORE_UNREAD,	/**< Flag whether "unread" icon is to be shown */
@@ -72,6 +76,49 @@ enum is_columns {
 	ITEMSTORE_LEN		/**< Number of columns in the itemstore */
 };
 
+typedef enum {
+	DEFAULT,
+	INTERNAL,
+	TAB,
+	EXTERNAL
+} open_link_target_type;
+
+static void
+launch_item_selected (open_link_target_type open_link_target)
+{
+	itemPtr		item;
+
+	item = itemlist_get_selected ();
+	if (item) {
+		gchar *link = item_make_link (item);
+
+		if (link) {
+			switch (open_link_target)
+			{
+			case DEFAULT:
+				itemview_launch_URL (link, FALSE);
+				break;
+			case INTERNAL:
+				itemview_launch_URL (link, TRUE);
+				break;
+			case TAB:
+				browser_tabs_add_new (link, link, FALSE);
+				break;
+			case EXTERNAL:
+				browser_launch_URL_external (link);
+				break;
+			}
+
+			g_free (link);
+		} else
+			ui_show_error_box (_("This item has no link specified!"));
+
+		item_unload (item);
+	} else {
+		liferea_shell_set_important_status_bar (_("No item has been selected"));
+	}
+}
+
 #define ITEM_LIST_VIEW_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), ITEM_LIST_VIEW_TYPE, ItemListViewPrivate))
 
 struct ItemListViewPrivate {
@@ -225,13 +272,65 @@ item_list_view_create_tree_store (void)
 	);
 }
 
+static void
+on_itemlist_selection_changed (GtkTreeSelection *selection, gpointer user_data)
+{
+	GtkTreeIter 	iter;
+	GtkTreeModel	*model;
+	itemPtr		item = NULL;
+
+	if (gtk_tree_selection_get_selected (selection, &model, &iter))
+		item = item_load (item_list_view_iter_to_id (ITEM_LIST_VIEW (user_data), &iter));
+
+	liferea_shell_update_item_menu (NULL != item);
+	if (item)
+		itemlist_selection_changed (item);
+}
+
+static void
+itemlist_sort_column_changed_cb (GtkTreeSortable *treesortable, gpointer user_data)
+{
+	gint		sortColumn, nodeSort;
+	GtkSortType	sortType;
+	gboolean	sorted, changed;
+	
+	if (feedlist_get_selected () == NULL)
+		return;
+	
+	sorted = gtk_tree_sortable_get_sort_column_id (treesortable, &sortColumn, &sortType);
+	if (!sorted)
+		return;
+		
+	switch (sortColumn) {
+		case IS_TIME:
+		default:
+			nodeSort = NODE_VIEW_SORT_BY_TIME;
+			break;
+		case IS_LABEL:
+			nodeSort = NODE_VIEW_SORT_BY_TITLE;
+			break;
+		case IS_STATE:
+			nodeSort = NODE_VIEW_SORT_BY_STATE;
+			break;
+		case IS_PARENT:
+		case IS_SOURCE:
+			nodeSort = NODE_VIEW_SORT_BY_PARENT;
+			break;
+	}
+
+	changed = node_set_sort_column (feedlist_get_selected (), nodeSort, sortType == GTK_SORT_DESCENDING);
+	if (changed)
+		feedlist_schedule_save ();
+}
+
 /**
  * Sets a GtkTreeView to the active GtkTreeView.
  */
 static void
 item_list_view_set_tree_store (ItemListView *ilv, GtkTreeStore *itemstore)
 {
-	GtkTreeModel    *model;
+	GtkTreeModel    	*model;
+	GtkTreeSelection	*select;
 
 	/* drop old tree store */
 	model = gtk_tree_view_get_model (ilv->priv->treeview);
@@ -245,7 +344,11 @@ item_list_view_set_tree_store (ItemListView *ilv, GtkTreeStore *itemstore)
 	
 	gtk_tree_view_set_model (ilv->priv->treeview, GTK_TREE_MODEL (itemstore));
 
-	item_list_view_prefocus (ilv);
+	/* Setup the selection handler */
+	select = gtk_tree_view_get_selection (ilv->priv->treeview);
+	gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
+	g_signal_connect (G_OBJECT (select), "changed",
+	                  G_CALLBACK (on_itemlist_selection_changed), ilv);
 }
 
 void
@@ -275,16 +378,21 @@ item_list_view_clear (ItemListView *ilv)
 {
 	GtkAdjustment		*adj;
 	GtkTreeStore		*itemstore;
+	GtkTreeSelection	*select;
 
 	itemstore = GTK_TREE_STORE (gtk_tree_view_get_model (ilv->priv->treeview));
-	
+
 	/* unselecting all items is important to remove items
 	   whose removal is deferred until unselecting */
-	gtk_tree_selection_unselect_all (gtk_tree_view_get_selection (ilv->priv->treeview));
+	select = gtk_tree_view_get_selection (ilv->priv->treeview);
+	gtk_tree_selection_unselect_all (select);
 	
-	adj = gtk_tree_view_get_vadjustment (ilv->priv->treeview);
+	adj = gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (ilv->priv->treeview));
 	gtk_adjustment_set_value (adj, 0.0);
-	gtk_tree_view_set_vadjustment (ilv->priv->treeview, adj);
+	gtk_scrollable_set_vadjustment (GTK_SCROLLABLE (ilv->priv->treeview), adj);
+
+	/* Disconnect signal handler to be safe */
+	g_signal_handlers_disconnect_by_func (G_OBJECT (select), G_CALLBACK (on_itemlist_selection_changed), ilv);
 
 	if (itemstore)
 		gtk_tree_store_clear (itemstore);
@@ -394,7 +502,8 @@ item_list_view_update (ItemListView *ilv, gboolean hasEnclosures)
 static gboolean
 on_item_list_view_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer data) 
 {
-	if ((event->type == GDK_KEY_PRESS) && (event->state == 0) && (event->keyval == GDK_Delete)) 
+	if ((event->type == GDK_KEY_PRESS) && (event->state == 0)
+	    && (event->keyval == GDK_KEY_Delete))
 		on_remove_item_activate(NULL, NULL);
 
 	return FALSE;
@@ -402,17 +511,22 @@ on_item_list_view_key_press_event (GtkWidget *widget, GdkEventKey *event, gpoint
 
 /* Show tooltip when headline's column text (IS_LABEL) is truncated. */
 
-static gint get_cell_renderer_width (GtkWidget *widget, GtkCellRenderer *cell, const gchar *text, gint weight)
+static gint
+get_cell_renderer_width (GtkWidget *widget, GtkCellRenderer *cell, const gchar *text, gint weight)
 {
-	PangoLayout *layout = gtk_widget_create_pango_layout (widget, text);
-	PangoAttrList *attrbs = pango_attr_list_new();
+	PangoLayout	*layout = gtk_widget_create_pango_layout (widget, text);
+	PangoAttrList	*attrbs = pango_attr_list_new();
+	PangoRectangle	rect;
+	gint		xpad = 0;
+
 	pango_attr_list_insert (attrbs, pango_attr_weight_new (weight));
 	pango_layout_set_attributes (layout, attrbs);
 	pango_attr_list_unref (attrbs);
-	PangoRectangle rect;
 	pango_layout_get_pixel_extents (layout, NULL, &rect);
 	g_object_unref (G_OBJECT (layout));
-	return (cell->xpad * 2) + rect.x + rect.width;
+
+	gtk_cell_renderer_get_padding (cell, &xpad, NULL);
+	return (xpad * 2) + rect.x + rect.width;
 }
 
 static gboolean
@@ -502,6 +616,9 @@ on_item_list_view_button_press_event (GtkWidget *treeview, GdkEventButton *event
 				ui_popup_item_menu (item, eb->button, eb->time);
 				result = TRUE;
 				break;
+			default:
+				/* Do nothing on buttons >= 4 */
+				break;
 		}
 		item_unload (item);
 	}
@@ -526,6 +643,15 @@ on_item_list_view_popup_menu (GtkWidget *widget, gpointer user_data)
 	return FALSE;
 }
 
+static void
+on_Itemlist_row_activated (GtkTreeView *treeview,
+                           GtkTreePath *path,
+			   GtkTreeViewColumn *column,
+			   gpointer user_data) 
+{
+    launch_item_selected (DEFAULT);
+}
+
 GtkTreeView *
 item_list_view_get_widget (ItemListView *ilv)
 {
@@ -545,7 +671,6 @@ item_list_view_create (GtkWidget *window)
 	ItemListView		*ilv;
 	GtkCellRenderer		*renderer;
 	GtkTreeViewColumn 	*column, *headline_column;
-	GtkTreeSelection	*select;
 	GtkWidget 		*ilscrolledwindow;
 
 	ilv = g_object_new (ITEM_LIST_VIEW_TYPE, NULL);
@@ -607,44 +732,10 @@ item_list_view_create (GtkWidget *window)
 
 	gtk_widget_set_has_tooltip (GTK_WIDGET (ilv->priv->treeview), TRUE);
 	g_signal_connect (G_OBJECT (ilv->priv->treeview), "query-tooltip", G_CALLBACK (on_item_list_view_query_tooltip), headline_column);
-
-	/* Setup the selection handler */
-	select = gtk_tree_view_get_selection (ilv->priv->treeview);
-	gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
-	g_signal_connect (G_OBJECT (select), "changed",
-	                  G_CALLBACK (on_itemlist_selection_changed), ilv);
 		  
 	return ilv;
 }
 
-/* typically called when filling the item tree view */
-void 
-item_list_view_prefocus (ItemListView *ilv)
-{
-	GtkWidget		*focus_widget;
-	GtkTreeSelection	*itemselection;
-	
-	/* the following is important to prevent setting the unread
-	   flag for the first item in the item list when the user does
-	   the first click into the treeview, if we don't do a focus and
-	   unselect, GTK would always (exception: clicking on first item)
-	   generate two selection-change events (one for the clicked and
-	   one for the selected item)!!! */
-
-	/* we need to restore the focus after we temporarily select the itemlist */
-	focus_widget = gtk_window_get_focus (GTK_WINDOW (liferea_shell_get_window ()));
-
-	/* prevent marking as unread before focussing, which leads to a selection */
-	gtk_widget_grab_focus (GTK_WIDGET (ilv->priv->treeview));
-
-	itemselection = gtk_tree_view_get_selection (ilv->priv->treeview);
-	if (itemselection)
-		gtk_tree_selection_unselect_all (itemselection);
-	
-	if (focus_widget)
-		gtk_widget_grab_focus (focus_widget);
-}
-
 static void
 item_list_view_add_item_to_tree_store (ItemListView *ilv, GtkTreeStore *itemstore, itemPtr item)
 {
@@ -712,52 +803,21 @@ item_list_view_enable_favicon_column (ItemListView *ilv, gboolean enabled)
 }
 
 void
-on_popup_launchitem_selected (void) 
+on_popup_launch_item_selected (void) 
 {
-	itemPtr		item;
-
-	item = itemlist_get_selected ();
-	if (item) {
-		gchar *link = item_make_link (item);
-
-		itemview_launch_URL (link, FALSE);
-
-		g_free (link);
-		item_unload (item);
-	} else {
-		liferea_shell_set_important_status_bar (_("No item has been selected"));
-	}
+    launch_item_selected (INTERNAL);
 }
 
 void
-on_popup_launchitem_in_tab_selected (void) 
+on_popup_launch_item_in_tab_selected (void) 
 {
-	itemPtr		item;
-	gchar           *link;
-
-	item = itemlist_get_selected ();
-	if (item) {
-		link = item_make_link (item);
-		if (link) {
-			browser_tabs_add_new (link, link, FALSE);
-			g_free (link);
-		} else
-			ui_show_error_box (_("This item has no link specified!"));
-			
-		item_unload (item);
-	} else {
-		liferea_shell_set_important_status_bar (_("No item has been selected"));
-	}
+    launch_item_selected (TAB);
 }
 
-void 
-on_Itemlist_row_activated (GtkTreeView *treeview,
-                           GtkTreePath *path,
-			   GtkTreeViewColumn *column,
-			   gpointer user_data) 
+void
+on_popup_launch_item_external_selected (void) 
 {
-
-	on_popup_launchitem_selected ();
+    launch_item_selected (EXTERNAL);
 }
 
 /* menu callbacks */
@@ -890,18 +950,6 @@ on_next_unread_item_activate (GtkMenuItem *menuitem, gpointer user_data)
 	itemlist_select_next_unread ();
 }
 
-void
-on_popup_next_unread_item_selected (gpointer callback_data, guint callback_action, GtkWidget *widget)
-{
-	itemlist_select_next_unread ();
-}
-
-void
-on_nextbtn_clicked (GtkButton *button, gpointer user_data)
-{
-	itemlist_select_next_unread ();
-}
-
 void
 on_popup_copy_URL_clipboard (void)
 {
@@ -934,54 +982,3 @@ on_popup_social_bm_item_selected (void)
 	else
 		liferea_shell_set_important_status_bar (_("No item has been selected"));
 }
-
-void
-on_itemlist_selection_changed (GtkTreeSelection *selection, gpointer user_data)
-{
-	GtkTreeIter 	iter;
-	GtkTreeModel	*model;
-	itemPtr		item = NULL;
-
-	if (gtk_tree_selection_get_selected (selection, &model, &iter))
-		item = item_load (item_list_view_iter_to_id (ITEM_LIST_VIEW (user_data), &iter));
-
-	liferea_shell_update_item_menu (NULL != item);
-	if (item)
-		itemlist_selection_changed (item);
-}
-
-void
-itemlist_sort_column_changed_cb (GtkTreeSortable *treesortable, gpointer user_data)
-{
-	gint		sortColumn, nodeSort;
-	GtkSortType	sortType;
-	gboolean	sorted, changed;
-	
-	if (feedlist_get_selected () == NULL)
-		return;
-	
-	sorted = gtk_tree_sortable_get_sort_column_id (treesortable, &sortColumn, &sortType);
-	if (!sorted)
-		return;
-		
-	switch (sortColumn) {
-		case IS_TIME:
-		default:
-			nodeSort = NODE_VIEW_SORT_BY_TIME;
-			break;
-		case IS_LABEL:
-			nodeSort = NODE_VIEW_SORT_BY_TITLE;
-			break;
-		case IS_STATE:
-			nodeSort = NODE_VIEW_SORT_BY_STATE;
-			break;
-		case IS_PARENT:
-		case IS_SOURCE:
-			nodeSort = NODE_VIEW_SORT_BY_PARENT;
-			break;
-	}
-
-	changed = node_set_sort_column (feedlist_get_selected (), nodeSort, sortType == GTK_SORT_DESCENDING);
-	if (changed)
-		feedlist_schedule_save ();
-}
