@@ -1,7 +1,7 @@
 /**
  * @file liferea_htmlview.c  Liferea embedded HTML rendering
  *
- * Copyright (C) 2003-2010 Lars Lindner <lars.lindner@gmail.com>
+ * Copyright (C) 2003-2012 Lars Windolf <lars.lindner@gmail.com>
  * Copyright (C) 2005-2006 Nathan J. Conrad <t98502@users.sourceforge.net> 
  * 
  * This program is free software; you can redistribute it and/or modify
@@ -26,6 +26,7 @@
 #include <glib.h>
 
 #include "browser.h"
+#include "browser_history.h"
 #include "comments.h"
 #include "common.h"
 #include "conf.h"
@@ -33,28 +34,43 @@
 #include "enclosure.h"
 #include "feed.h"
 #include "feedlist.h"
-#include "itemlist.h"
 #include "net.h"
 #include "net_monitor.h"
 #include "social.h"
 #include "render.h"
 #include "htmlview.h"
 #include "ui/browser_tabs.h"
-#include "ui/liferea_shell.h"
 #include "ui/item_list_view.h"
-#include "ui/ui_common.h"
-#include "ui/ui_prefs.h"
+
+/* The LifereaHtmlView is a complex widget used to present both internally
+   rendered content as well as serving as a browser widget. It automatically
+   switches on a toolbar for history and URL navigation when browsing 
+   external content.
+
+   When serving internal content it reacts to different internal link schemata
+   to trigger functionality inside Liferea. To avoid websites hijacking this we
+   keep a flag to support the link schema only on liferea_htmlview_write()
+ */
+ 
 
 #define RENDERER(htmlview)	(htmlview->priv->impl)
 
 #define LIFEREA_HTMLVIEW_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), LIFEREA_HTMLVIEW_TYPE, LifereaHtmlViewPrivate))
 
 struct LifereaHtmlViewPrivate {
-	GtkWidget	*renderWidget;
+	GtkWidget	*renderWidget;		/**< The HTML widget (e.g. Webkit widget) */
+	GtkWidget	*container;		/**< Outer container including render widget and toolbar */
+	GtkWidget	*toolbar;		/**< The navigation toolbar */
+
+	GtkWidget	*forward;		/**< The forward button */
+	GtkWidget	*back;			/**< The back button */
+	GtkWidget	*urlentry;		/**< The URL entry widget */
+	browserHistory	*history;		/**< The browser history */
+
 	gboolean	internal;		/**< TRUE if internal view presenting generated HTML with special links */
 	gboolean	forceInternalBrowsing;	/**< TRUE if clicked links should be force loaded within this view (regardless of global preference) */
 	
-	htmlviewImplPtr impl;			/**< browser widget support implementation */
+	htmlviewImplPtr impl;			/**< Browser widget support implementation */
 };
 
 enum {
@@ -64,19 +80,68 @@ enum {
 	LAST_SIGNAL
 };
 
+/* LifereaHtmlView toolbar callbacks */
+
+static gboolean
+on_htmlview_url_entry_activate (GtkWidget *widget, gpointer user_data)
+{
+	LifereaHtmlView	*htmlview = LIFEREA_HTMLVIEW (user_data);
+	gchar	*url;
+
+	url = (gchar *)gtk_entry_get_text (GTK_ENTRY (widget));
+	liferea_htmlview_launch_URL_internal (htmlview, url);
+
+	return TRUE;
+}
+
+static void
+on_htmlview_history_back (GtkWidget *widget, gpointer user_data)
+{
+	LifereaHtmlView	*htmlview = LIFEREA_HTMLVIEW (user_data);
+	gchar		*url;
+
+	url = browser_history_back (htmlview->priv->history);
+
+	gtk_widget_set_sensitive (htmlview->priv->forward, browser_history_can_go_forward (htmlview->priv->history));
+	gtk_widget_set_sensitive (htmlview->priv->back,    browser_history_can_go_back (htmlview->priv->history));
+
+	liferea_htmlview_launch_URL_internal (htmlview, url);
+	gtk_entry_set_text (GTK_ENTRY (htmlview->priv->urlentry), url);
+}
+
+static void
+on_htmlview_history_forward (GtkWidget *widget, gpointer user_data)
+{
+	LifereaHtmlView	*htmlview = LIFEREA_HTMLVIEW (user_data);
+	gchar		*url;
+
+	url = browser_history_forward (htmlview->priv->history);
+
+	gtk_widget_set_sensitive (htmlview->priv->forward, browser_history_can_go_forward (htmlview->priv->history));
+	gtk_widget_set_sensitive (htmlview->priv->back,    browser_history_can_go_back (htmlview->priv->history));
+
+	liferea_htmlview_launch_URL_internal (htmlview, url);
+	gtk_entry_set_text (GTK_ENTRY (htmlview->priv->urlentry), url);
+}
+
+
+/* LifereaHtmlView class */
+
 static guint liferea_htmlview_signals[LAST_SIGNAL] = { 0 };
 
 static GObjectClass *parent_class = NULL;
 
-/* -------------------------------------------------------------------- */
-/* Liferea HTML rendering object					*/
-/* -------------------------------------------------------------------- */
-
 G_DEFINE_TYPE (LifereaHtmlView, liferea_htmlview, G_TYPE_OBJECT);
 
 static void
 liferea_htmlview_finalize (GObject *object)
 {
+	LifereaHtmlView *htmlview = LIFEREA_HTMLVIEW (object);
+
+	browser_history_free (htmlview->priv->history);
+
+	g_signal_handlers_disconnect_by_data (network_monitor_get (), object);
+
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -133,10 +198,45 @@ liferea_htmlview_class_init (LifereaHtmlViewClass *klass)
 static void
 liferea_htmlview_init (LifereaHtmlView *htmlview)
 {
+	GtkWidget *widget, *image;
+
 	htmlview->priv = LIFEREA_HTMLVIEW_GET_PRIVATE (htmlview);
 	htmlview->priv->internal = FALSE;
 	htmlview->priv->impl = htmlview_get_impl ();
 	htmlview->priv->renderWidget = RENDERER (htmlview)->create (htmlview);
+	htmlview->priv->container = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
+	htmlview->priv->history = browser_history_new ();
+	htmlview->priv->toolbar = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
+	
+	widget = gtk_button_new ();	
+	gtk_button_set_relief (GTK_BUTTON (widget), GTK_RELIEF_NONE);
+	image = gtk_image_new_from_stock ("gtk-go-back", GTK_ICON_SIZE_BUTTON);
+	gtk_widget_show (image);
+	gtk_container_add (GTK_CONTAINER (widget), image);
+	gtk_box_pack_start (GTK_BOX (htmlview->priv->toolbar), widget, FALSE, FALSE, 0);
+	g_signal_connect ((gpointer)widget, "clicked", G_CALLBACK (on_htmlview_history_back), (gpointer)htmlview);
+	gtk_widget_set_sensitive (widget, FALSE);
+	htmlview->priv->back = widget;
+
+	widget = gtk_button_new ();
+	gtk_button_set_relief (GTK_BUTTON(widget), GTK_RELIEF_NONE);
+	image = gtk_image_new_from_stock ("gtk-go-forward", GTK_ICON_SIZE_BUTTON);
+	gtk_widget_show (image);
+	gtk_container_add (GTK_CONTAINER (widget), image);
+	gtk_box_pack_start (GTK_BOX (htmlview->priv->toolbar), widget, FALSE, FALSE, 0);
+	g_signal_connect ((gpointer)widget, "clicked", G_CALLBACK (on_htmlview_history_forward), (gpointer)htmlview);
+	gtk_widget_set_sensitive (widget, FALSE);
+	htmlview->priv->forward = widget;
+
+	widget = gtk_entry_new ();
+	gtk_box_pack_start (GTK_BOX (htmlview->priv->toolbar), widget, TRUE, TRUE, 0);
+	g_signal_connect ((gpointer)widget, "activate", G_CALLBACK (on_htmlview_url_entry_activate), (gpointer)htmlview);
+	htmlview->priv->urlentry = widget;
+
+	gtk_box_pack_start (GTK_BOX (htmlview->priv->container), htmlview->priv->toolbar, FALSE, FALSE, 0);
+	gtk_box_pack_end (GTK_BOX (htmlview->priv->container), htmlview->priv->renderWidget, TRUE, TRUE, 0);
+
+	gtk_widget_show_all (htmlview->priv->container);
 }
 
 static void
@@ -194,13 +294,16 @@ liferea_htmlview_new (gboolean forceInternalBrowsing)
 GtkWidget *
 liferea_htmlview_get_widget (LifereaHtmlView *htmlview)
 {
-	return htmlview->priv->renderWidget;
+	return htmlview->priv->container;
 }
 
 void
 liferea_htmlview_write (LifereaHtmlView *htmlview, const gchar *string, const gchar *base)
 { 
 	const gchar	*baseURL = base;
+
+	if (!htmlview)
+		return;
 	
 	htmlview->priv->internal = TRUE;	/* enables special links */
 	
@@ -219,6 +322,9 @@ liferea_htmlview_write (LifereaHtmlView *htmlview, const gchar *string, const gc
 	} else {
 		(RENDERER (htmlview)->write) (htmlview->priv->renderWidget, string, strlen (string), baseURL, "application/xhtml+xml");
 	}
+
+	/* We hide the toolbar as it should only be shown when loading external content */
+	gtk_widget_hide (htmlview->priv->toolbar);
 }
 
 void
@@ -233,7 +339,7 @@ liferea_htmlview_clear (LifereaHtmlView *htmlview)
 	g_string_free (buffer, TRUE);
 }
 
-gboolean
+static gboolean
 liferea_htmlview_is_special_url (const gchar *url)
 {
 	/* match against all special protocols, simple
@@ -250,8 +356,6 @@ struct internalUriType {
 };
 
 static struct internalUriType internalUriTypes[] = {
-	{ "flag",		itemlist_toggle_flag },
-	{ "bookmark",		social_add_bookmark },
 	{ "refresh-comments",	comments_refresh },
 	{ NULL,			NULL }
 };
@@ -272,6 +376,18 @@ liferea_htmlview_title_changed (LifereaHtmlView *htmlview, const gchar *title)
 void
 liferea_htmlview_location_changed (LifereaHtmlView *htmlview, const gchar *location)
 {
+	if (!htmlview->priv->internal) {
+		browser_history_add_location (htmlview->priv->history, location);
+
+		gtk_widget_set_sensitive (htmlview->priv->forward, browser_history_can_go_forward (htmlview->priv->history));
+		gtk_widget_set_sensitive (htmlview->priv->back,    browser_history_can_go_back (htmlview->priv->history));
+
+		gtk_entry_set_text (GTK_ENTRY (htmlview->priv->urlentry), location);
+
+		/* We show the toolbar as it should be visible when loading external content */
+		gtk_widget_show_all (htmlview->priv->toolbar);
+	}
+
 	g_signal_emit_by_name (htmlview, "location-changed", location);
 }
 
@@ -348,6 +464,13 @@ liferea_htmlview_launch_URL_internal (LifereaHtmlView *htmlview, const gchar *ur
 {
 	/* before loading untrusted URLs suppress internal link schema */
 	htmlview->priv->internal = FALSE;
+
+	browser_history_add_location (htmlview->priv->history, (gchar *)url);
+
+	gtk_widget_set_sensitive (htmlview->priv->forward, browser_history_can_go_forward (htmlview->priv->history));
+	gtk_widget_set_sensitive (htmlview->priv->back,    browser_history_can_go_back (htmlview->priv->history));
+
+	gtk_entry_set_text (GTK_ENTRY (htmlview->priv->urlentry), url);
 	
 	(RENDERER (htmlview)->launch) (htmlview->priv->renderWidget, url);
 }
@@ -382,6 +505,12 @@ liferea_htmlview_do_zoom (LifereaHtmlView *htmlview, gboolean in)
 
 static void
 on_popup_launch_link_activate (GtkWidget *widget, gpointer user_data)
+{
+	itemview_launch_URL ((gchar *)user_data, TRUE /* use internal browser */);
+}
+
+static void
+on_popup_launch_link_external_activate (GtkWidget *widget, gpointer user_data)
 {
 	browser_launch_URL_external ((gchar *)user_data);
 }
@@ -407,54 +536,10 @@ on_popup_copy_url_activate (GtkWidget *widget, gpointer user_data)
 	g_free (link);
 }
 
-static void
-on_save_url (const gchar *filename, gpointer user_data)
-{
-	/* FIXME: The following partially duplicates download code in enclosure.c! */
-	enclosureDownloadToolPtr tool;
-	gchar	*stdout_message = NULL, *stderr_message = NULL;
-	GError	*error = NULL;
-	gint	status = 0;
-	gchar	*uriQ, *uri = (gpointer)user_data;
-	gchar	*cmd;
-	
-	if (!filename)
-		return;
-
-	tool = prefs_get_download_tool ();
-	uriQ = g_shell_quote (uri);
-	cmd = g_strdup_printf (tool->format, filename, uriQ);
-
-	debug1 (DEBUG_UPDATE, "running download command \"%s\"", cmd);
-	g_spawn_command_line_sync (cmd, &stdout_message, &stderr_message, &status, &error);
-	
-	if ((error && (0 != error->code)) || !WIFEXITED(status) || WEXITSTATUS(status)) {
-		g_warning ("Failed to execute command \"%s\", exited: %i, status: %i, stderr: %s, stdout: %s", cmd, WIFEXITED(status), WEXITSTATUS(status), stderr_message?:"", stdout_message?:"");
-		liferea_shell_set_status_bar (_("Download FAILED: \"%s\""), uri);
-	} else {
-		liferea_shell_set_status_bar (_("Download finished."));
-	}
-	
-	if (error)
-		g_error_free (error);
-	g_free (stdout_message);
-	g_free (stderr_message);
-	g_free (cmd);
-}
-
 static void
 on_popup_save_url_activate (GtkWidget *widget, gpointer user_data)
 {
-	gchar	*uri = (gpointer)user_data;
-	gchar	*filename;
-	
-	filename = strrchr (uri, '/');
-	if (filename)
-		filename++; /* Skip the slash to find the filename */
-	else
-		filename = uri;
-	
-	ui_choose_file (_("Choose File"), GTK_STOCK_SAVE_AS, TRUE, on_save_url, NULL, filename, NULL, NULL, uri);
+	enclosure_download (NULL, (const gchar *)user_data, TRUE);
 }
 
 static void
@@ -539,8 +624,9 @@ liferea_htmlview_prepare_context_menu (LifereaHtmlView *htmlview, GtkMenu *menu,
 	/* and now add all we want to see */
 	if (link) {
 		gchar *path;
-		menu_add_option (menu, _("Launch Link In _Tab"), NULL, G_CALLBACK (on_popup_open_link_in_tab_activate), (gpointer)linkUri);
-		menu_add_option (menu, _("_Launch Link In Browser"), NULL, G_CALLBACK (on_popup_launch_link_activate), (gpointer)linkUri);
+		menu_add_option (menu, _("Open Link In _Tab"), NULL, G_CALLBACK (on_popup_open_link_in_tab_activate), (gpointer)linkUri);
+		menu_add_option (menu, _("_Open Link In Browser"), NULL, G_CALLBACK (on_popup_launch_link_activate), (gpointer)linkUri);
+		menu_add_option (menu, _("_Open Link In External Browser"), NULL, G_CALLBACK (on_popup_launch_link_external_activate), (gpointer)linkUri);
 		menu_add_separator (menu);
 		
 		path = g_strdup_printf (_("_Bookmark Link at %s"), social_get_bookmark_site ());
